/******************************************************/
/* File: queue.c                                      */
/* used in the Dynamic Calendar Queue Data Structure  */
/* described by Brown (1988)						  */                     
/* note: this file is not to used with the Basic Q    */
/* Last Updated July 30/2001 by Philip Henderson      */
/******************************************************/
#include "queue.h"


int BASIC = 2000;

/****************************************************/
/* printQ											*/
/* prints out the entire queue with all timestamps  */
/* to standard output								*/
/****************************************************/

void printQ(calendarQ* Q){

Bucket* iBucket;

node* thisNode;

 int i, j;

printf("\n\n===CalendarQ Stats===========================\n");

printf("qsize = %d, nBuckets = %d, firstsub = %d \n", 
	   Q->qsize, Q->nBuckets, Q->firstsub);
printf("top_thresh = %d, bot_thresh = %d \n", Q->top_threshold, Q->bot_threshold);
printf("BucketWidth = %lf \n", Q->BucketWidth);
printf("=============================================\n\n");



 printf("data in buckets: \n contIdx bcktIdx | data \n ------- -------   ----\n");

 for (i = 0; i < Q->nBuckets; i++){
 iBucket = Q->BcktContainer[i + Q->firstsub];
	 printf("[%d] (%d)  [%d]  ", Q->firstsub + i, iBucket->nodeCount, i);

	 if(iBucket->nodeCount > 0){
		thisNode = iBucket->head;
		 printf(" %lf, " , thisNode->Prio);
		for(j = 1U; j< iBucket->nodeCount; j++){
			thisNode = thisNode->next;	
			printf(" %lf, ", thisNode->Prio);
		}
	 }
	 printf("\n");
 }
}


/****************************************************/
/* fprintQ											*/
/* prints out the entire queue with all timestamps  */
/* to the file specified							*/
/****************************************************/
void fprintQ(calendarQ* Q, FILE *fp ){

Bucket* iBucket;

node* thisNode;

int i, j;

fprintf(fp, "\n\n===CalendarQ Stats===========================\n");

fprintf(fp, "qsize = %d, nBuckets = %d, firstsub = %d \n", 
	   Q->qsize, Q->nBuckets, Q->firstsub);
fprintf(fp, "top_thresh = %d, bot_thresh = %d \n", Q->top_threshold, Q->bot_threshold);
fprintf(fp, "BucketWidth = %lf \n", Q->BucketWidth);
fprintf(fp, "=============================================\n\n");



 fprintf(fp, "data in buckets: \n contIdx bcktIdx | data \n ------- -------   ----\n");

 for (i = 0; i < Q->nBuckets; i++){
 iBucket = Q->BcktContainer[i + Q->firstsub];
	 fprintf(fp, "[%d] (%d)  [%d]  ", Q->firstsub + i, iBucket->nodeCount, i );

	 if(iBucket->nodeCount > 0){
		thisNode = iBucket->head;
		 fprintf(fp, " %lf, " , thisNode->Prio);
		for(j = 1; j< iBucket->nodeCount; j++){
			thisNode = thisNode->next;	
			fprintf(fp, " %lf, ", thisNode->Prio);
		}
	 }
	 fprintf(fp, "\n");
 }
}



/****************************************************/
/* InitQueue										*/
/* initializes the queue with the necessary memory  */
/*                      							*/
/****************************************************/
calendarQ* InitQueue()
{
  calendarQ* Q;
  int i;

  /*initialize memory for a new calendarQ */
  Q = (calendarQ *)malloc(sizeof(calendarQ));
 
  /* initialize memory for the BcktContainer */
 /* myCalendarQ->BcktContainer = (Bucket *)calloc(QSPACE, sizeof(Bucket));*/
  Q->resizeEnabled = TRUE;
  /*initialize the maximum number of buckets in the bcktContainer*/
  for(i = 0; i < QSPACE; i++){
   Q->BcktContainer[i] = createBucket(0.00, 0.00);
  }
  Q->localinitCounter = 0;
  LocalInit(Q, 0, 2, 1.0, 0.0);

	Q->localinitCounter = 1;
	Q->resizeCount = 0;
  return Q;

}

/****************************************************/
/* createBucket										*/
/* initializes a bucket with the necessary memory  */
/*                      							*/
/****************************************************/

Bucket* createBucket(double lBound, double uBound){
Bucket* myBucket;

myBucket = (Bucket*)malloc(sizeof(Bucket));
myBucket->head = NULL;
myBucket->tail = NULL;
myBucket->nodeCount = 0;
myBucket->BotPrio = 0;
myBucket->TopPrio = 0;

return myBucket;
}


/****************************************************/
/* LocalInit										*/
/* initializes a queue subarray within the larger   */
/* bucket array at the beginning of each resize     */
/* (as described by Brown88)						*/
/****************************************************/
void LocalInit(calendarQ* Q, int qbase, 
	       int nBuck, double bWidth, double startprio){
  
  long int n;
  double uBound, lBound;
   lBound = 0;
   uBound = lBound + bWidth;
Q->localinitCounter++;
  Q->firstsub = qbase;

  /* pointer to start of bucketArray[] in BcktContainer[] */
 /* Q->BucketArray = Q->BcktContainer[qbase];*/
  Q->qsize = 0;
  Q->BucketWidth = bWidth;
  Q->nBuckets = nBuck;

  Q->LastPrio = startprio;

  n = (long)(startprio/bWidth);
  Q->LastBucket = n % (Q->nBuckets);
  Q->BucketTop = (n + 1) * Q->BucketWidth + 0.5 * Q->BucketWidth;
  
  Q->bot_threshold = (Q->nBuckets/2) - 2;
  Q->top_threshold = 2 * (Q->nBuckets);
}



/***************************************************************************/
/*                                                                         */
/* resize:                                                                 */
/*     This routine copies the queue into a calendar with the number of    */
/* buckets specified in the parameter newsize.  The new queue is in the    */
/* opposite end of calq from the original.                                 */
/* Q: a pointer to a calendarQ object									   */
/*   newBucketCount: 2 * the current bucket count for a basic resize,	   */
/*                    otherwise, the current bucket count for a			   */
/*					redistribution type (DCQ) resize					   */
/***************************************************************************/

void resize(calendarQ* Q,  int newBucketCount){
  int i, j, oldnbuckets, initNodeCount, ArrayIndex;
  node* thisNode;
  int oldFirstSub;

  Bucket* oldbucket;
  Bucket* iBucket;
  
  if (!Q->resizeEnabled) return;
	Q->resizeCount++;
   Q->BucketWidth = newwidth(Q);
	
  /* Save location and size of old calendar for use when copying calendar*/
  oldbucket = Q->BcktContainer[Q->firstsub];
  oldnbuckets = Q->nBuckets;
  oldFirstSub = Q->firstsub;

  /*initialize new calendar*/
  if (Q->firstsub == 0){
      LocalInit(Q, QSPACE-newBucketCount, newBucketCount,
		Q->BucketWidth, Q->LastPrio);
    }
  else{
    LocalInit(Q, 0, newBucketCount, Q->BucketWidth, Q->LastPrio);
  }

  /*Copy queue elements to new calendar. */
  for (i = oldnbuckets - 1; i>=0; --i)
 
    /*transfer elements from bucket i to new calendar by enqueuing them*/
    {   
       	ArrayIndex = oldFirstSub + i;
    	iBucket = Q->BcktContainer[ArrayIndex];

	if (iBucket->nodeCount > 0){
		  thisNode = iBucket->head;
		  initNodeCount = iBucket->nodeCount;
    		
    		for ( j = 0 ; j< initNodeCount; j++){

				thisNode = iBucket->head;

				if (iBucket->nodeCount > 1){
      				iBucket->head = thisNode->next;
					iBucket->head->prev = NULL;

					iBucket->TopPrio = iBucket->head->Prio;
				}
				else{
					iBucket->head = NULL;
					iBucket->TopPrio = 0;
					iBucket->BotPrio = 0;
					iBucket->tail = 0;
				}
      			thisNode->next = NULL;
      			thisNode->prev = NULL;
      			iBucket->nodeCount--;
      			enqueue(Q, thisNode);
      		}
      	}
      }
}

/*************************************************************************/
/*                                                                       */
/* newwidth:                                                             */
/*     This function returns the width that the buckets should have      */
/* based on a random sample of the queue                                 */
/*                                                                       */
/*************************************************************************/

double newwidth(calendarQ* Q)
{
  /* This calculates the width to use for buckets */
  int nSamples, count,tempLastBucket, i;
  double tempLastPrio,  tempBucketTop, 
	  SeparationSum,  AverageSeparation, newAverage;

  /*the Separation distance is the difference of priorities
    between two adjacent nodes in a list.*/
  node* temp[25];

  SeparationSum = 0;

  /*Decide how many queue elements to sample */
  if(Q->qsize < 2) return (1.0);
  if(Q->qsize <= 5) nSamples = Q->qsize;
  else nSamples = 5 + (int)((float)Q->qsize/10);
  if(nSamples > 25) nSamples = 25;

  /*record Q variables*/
  tempLastPrio = Q->LastPrio;
  tempBucketTop = Q->BucketTop;
  tempLastBucket = Q->LastBucket;

  Q->resizeEnabled = FALSE;
  for(i = 0; i< nSamples; i++){
	  temp[i] = dequeue(Q);
	  if (i > 0) SeparationSum+= (temp[i]->Prio - temp[i-1]->Prio);
  }
  AverageSeparation = SeparationSum/(float)(nSamples-1);

  /*recalculate average*/
  SeparationSum = 0;   /*re-initialize SeparationSum*/
  count = 0;
  
  for (i = 1; i < nSamples; i++){
	  if((temp[i]->Prio - temp[i-1]->Prio) < AverageSeparation * 2){
		  SeparationSum += (temp[i]->Prio - temp[i-1]->Prio);
		count++;
	  }
  }
  	newAverage = SeparationSum/count;
  	/*now enqueue items that were dequeued for this test*/
	for (i = 0; i< nSamples; i++){
  		enqueue(Q, temp[i]);
  		}
  	/*restore saved values*/
  Q->LastPrio =tempLastPrio;
  Q->BucketTop = tempBucketTop;
  Q->LastBucket = tempLastBucket;
  Q->resizeEnabled = TRUE;

  	return (newAverage * Q->widthFactor); 

}

node* createNode(double priority){
node* newNode;

/*create a new node*/
  newNode = (node *)malloc(sizeof(node));
  newNode->Prio = priority;
  
return newNode;
}




node* dequeue(calendarQ* Q)
{
  /* this will remove the node with the lowest priority from 
     the queue and return a pointer to the node containing it.*/

  double minPrio, TempPrio;
  int i, ArrayIndex;
  node* currNode;
  Bucket* iBucket;
  

  /*initialize minPrio*/

  minPrio = 100000;

  if (Q->qsize == 0) return(NULL);
  

  /*search one year of buckets*/
  /*i ranges from 0 to nBuckets*/
  for (i = Q->LastBucket ; ;) 
  {		  /*check bucket i */
	ArrayIndex = i + Q->firstsub;
	iBucket = Q->BcktContainer[ArrayIndex];

	if((iBucket->nodeCount != 0) && (iBucket->head->Prio < minPrio))
		minPrio = iBucket->head->Prio;


    if(iBucket->nodeCount != 0 &&
		iBucket->TopPrio < Q->BucketTop)
	{
	  /*item has been found (it is the top of the list
		  for this bucket) -remove it from the list*/
	
		currNode = iBucket->head;
		if(iBucket->nodeCount > 1){
			iBucket->head = iBucket->head->next;
			iBucket->head->prev = NULL;
			iBucket->TopPrio = iBucket->head->Prio;
		}
		else{
			iBucket->head = NULL;
			iBucket->TopPrio = 0;
		}
		currNode->next = NULL;
		currNode->prev = NULL;
		iBucket->nodeCount--;
		
	
	  /*update position on calendar*/
	  Q->LastBucket = i;
	  Q->LastPrio = currNode->Prio;
	  Q->qsize--;

	  /*Halve calendar size if needed */
	  if (Q->qsize < Q->bot_threshold)
	    resize(Q, Q->nBuckets/2);
  
	  return currNode;
	}
      else
	{  /* prepare to check next bucket
	      or else go to a direct search*/
	  i++;
	  if(i == Q->nBuckets) i = 0; /*loop around to the next year*/
	  Q->BucketTop += Q->BucketWidth;
	  

	  if(Q->BcktContainer[i + Q->firstsub]->nodeCount > 0){
	  TempPrio = Q->BcktContainer[i + Q->firstsub]->head->Prio;
	  /*record highest priority item to date in case it is needed
	    later in a direct search*/
	  if (TempPrio < minPrio) minPrio = TempPrio;
	  }

	  if(i == Q->LastBucket) break;
	  /*lowest priority event is not in this year*/

	  /*Go to a direct search*/
	  }
    }

  /*we have the highest priority from previous looping*/
  /*find the bucket with that element*/


   i = (int)(minPrio/Q->BucketWidth); /* find virtual bucket.*/
   i = i% Q->nBuckets;         /* find actual bucket*/
   iBucket = Q->BcktContainer[i + Q->firstsub];

   currNode = iBucket->head;
   if (iBucket->nodeCount > 1){
     iBucket->head = iBucket->head->next;
     iBucket->head->prev = NULL;
     iBucket->TopPrio = iBucket->head->Prio;
   }
   else{
	 iBucket->tail = NULL;
	 iBucket->head = NULL;
     iBucket->TopPrio = 0;
	 iBucket->BotPrio = 0;
   }

   iBucket->nodeCount--;
   Q->qsize--;

   currNode->next = NULL;
   currNode->prev = NULL;
   Q->LastBucket = i;
   Q->LastPrio = currNode->Prio;
   Q->BucketTop = (i + 1) * Q->BucketWidth + 0.5 * Q->BucketWidth;

   if(Q->qsize < Q->bot_threshold)
     resize(Q, Q->nBuckets/2);
   
   return currNode;

}

void enqueue(calendarQ* Q, node* entry)
{/*this adds one entry to the queue*/

  Bucket* iBucket;
  node* currNode;
  double priority;
	int EnqCnt = 0;
  int i;

  priority = entry->Prio;
 
  /*calculate the number of the bucket in which to place the new node*/
  /*this is equivalent to hashing*/
  i = (int)(priority/Q->BucketWidth); /* find virtual bucket.*/
  i = i% Q->nBuckets;         /* find actual bucket*/

  /*firstsub is the offset*/
  iBucket = Q->BcktContainer[i+ Q->firstsub];
  /*insert entry into bucketarray i in sorted list*/
  
 if (iBucket->nodeCount == 0){      /*first node in this bucket*/
   iBucket->head = entry;
   entry->prev = NULL;
   entry->next = NULL;
   iBucket->TopPrio = priority;
   iBucket->tail = entry;
   iBucket->BotPrio = priority;
   }
 else{                          /*search for position to put new node*/
   if(priority < iBucket->TopPrio) 
/*This is the highest priority in the list */
     { entry->next = iBucket->head;
     iBucket->head->prev = entry;
     entry->prev = NULL;
     iBucket->head = entry;
     iBucket->TopPrio = priority;
     }
   else if(priority >= iBucket->BotPrio){
   /*this is the lowest priority item in the list*/
     entry->prev = iBucket->tail;
     iBucket->tail->next = entry;
     entry->next = NULL;
     iBucket->tail = entry;
     iBucket->BotPrio = priority;
   }
   else{                /*priority is between the TopPrio and BotPrio*/
     currNode = iBucket->head;
	EnqCnt++;
     /*put new node after all nodes with higher priority, or
       if all nodes have same priority, put new node after them*/
     
     while((priority >= currNode->Prio) && (EnqCnt < iBucket->nodeCount)){
	 /*linear search (could be converted to a binary search)*/
     EnqCnt++;
       currNode = currNode->next; 
     }
	 if(priority != currNode->Prio){
		/*put entry before currNode*/
		entry->next = currNode;
		entry->prev = currNode->prev;
		currNode->prev->next = entry;
		currNode->prev = entry;
	 }else{
		 /*put duplicate entries in fifo order*/
		 entry->next = currNode->next;
		 entry->prev = currNode;
		 currNode->next = entry;
		 if (currNode != iBucket->tail){
			 currNode->next->prev = entry;
		 }else{
			 iBucket->tail = entry;
			 iBucket->BotPrio = priority;
			 entry->next = NULL;
		 }
	 }
   }
 }
 iBucket->nodeCount++;
 
   Q->qsize++;

   if(Q->qsize > Q->top_threshold){
  resize(Q, 2* Q->nBuckets);
    }	
 }


