/*WinQDriver.c*/
/*note: this driver will only work with VisualC++ compiler*/
/*driver program to test the Dynamic CalendarQ*/
/*Created June 1st, 2001 by Philip Henderson*/
/*last updated July 30/2001*/

#include <windows.h>

#include "queue.h"
#include "camel.h"

#define MAXPACKETCOUNT 40000
FILE *fp;
node* packetList[MAXPACKETCOUNT];
double RandomNumberList[MAXPACKETCOUNT];

void createPackets(unsigned int n,  Distribution* myDist){
  /*n is the number of packets to create*/
  
  unsigned int i;
  
  if (n > MAXPACKETCOUNT){
    printf("can't create more than %d packets.\n", MAXPACKETCOUNT);
    return;
  }

    for (i = 0; i < n; i++){
      packetList[i] = createNode(myDist->dataList[i]);
    }
  
  return;
}



void KillCalendar(calendarQ* Q){
/*frees all memory allocated for a calendarQ*/
	Bucket* iBucket;
	node* currNode;
	int nodeSum = 0;

	int i;

 for (i = 0; i < Q->nBuckets; i++){
 iBucket = Q->BcktContainer[i + Q->firstsub];
	  if(iBucket->nodeCount > 0){
	
		 
		while(iBucket->nodeCount > 1){
			currNode = iBucket->tail;
			iBucket->tail = currNode->prev;
		
			currNode->prev = NULL;
			free(currNode);
			iBucket->nodeCount = iBucket->nodeCount - 1;
		}
		/*free the head node*/
		if(iBucket->nodeCount == 1){
			currNode = iBucket->head;
			iBucket->head = NULL;
			iBucket->tail = NULL;
			currNode->next = NULL;
			free(currNode);
			iBucket->nodeCount = 0;	
		}
	}
 }
 for(i = 0; i < QSPACE; i++){
	 free(Q->BcktContainer[i]);
 }

	free(Q);
}

void PrintEventDistribution(calendarQ* Q, int k, char* filename){
	Bucket* iBucket;
	int i, EventCount;
	FILE* fp2;
	 fp2 = fopen(filename, "w");

		fprintf(fp2,"queue size %d\n", Q->qsize);
		fprintf(fp2, "Event distribution for this calendarQ:\n");
		fprintf(fp2, "BucketIndex     EventCount\n");
		fprintf(fp2, "____________    ___________\n");
		
	for (i = 0; i < Q->nBuckets; i++){
		iBucket = Q->BcktContainer[Q->firstsub + i];
		EventCount = iBucket->nodeCount;
	
			fprintf(fp2, " %d               %d\n", i, EventCount);
	
	}
	fclose(fp2);
}


int main()
{
	/*usage: WinQDriver bucketWidthMultiplicationFactor*/

	calendarQ* myCalendarQ;
	int w,i, j,f, n,k, choice, nHOLD, alpha, FLAG, next, maxQSize, maxSeconds;
	int delta;
	FILE* fpDebug;
	Distribution* myDist;
	Random myRand;

	RNG myRNG(RNG::RAW_SEED_SOURCE);
	myRNG.default_ = &myRNG;

	LARGE_INTEGER liStart, liEnd, liPerfFreq;
	node* myNode;
	double  duration, AverageHoldTime, durationPerQOperation;
	char* DistType;
	char* filename;
/*	FILE* DistFile;
	DistFile = fopen("DistributionData.txt","w");*/
	fp = fopen("output.txt", "w");
	w = 3;
	printf("width factor = %d \n", w);
	choice = 0;
	n= MAXPACKETCOUNT;
	delta = 1000;
	nHOLD = 10000;
	FLAG = 0;

	maxQSize = 20000;
	QueryPerformanceFrequency(&liPerfFreq);
	maxSeconds = 600; /*the total amount of simulated time*/
	myDist = createDistribution();
	
/*Create an array of pseudo-random numbers to use as input*/

 for (k = 1; k <= 6; k++){
	 alpha = delta;
	switch(k)

	{

	case 1:
		DistType = "Positive Triangular (0,1,1)";
		filename = "PosTriangle.txt";
		break;
	case 2:
		DistType = "Negative Triangular (0,1,0)";
		filename = "NegTriangle.txt";
		break;
	case 3:
		DistType = "Rectangular Distribution";
		filename = "Rectangle.txt";
		break;
	case 4:
		DistType = "Jitter Distribution";
		filename = "Jitter.txt";
		break;
	case 5:
		DistType = "Camel Distribution(0.8, 0.2)";
		filename = "Camel82.txt";
		break;
	case 6:
		DistType = "Camel Distribution(0.99, 0.01)";
		filename = "Camel9901.txt";
		break;
	}

	fprintf(fp, "\nDistribution: = %s \n", DistType);
	fprintf(fp,"\nUpDown model: time for enqueue and dequeue of N events\n");
	fprintf(fp, "QSize   UpDown Duration (seconds)\n");
	fprintf(fp, "______  __________________________________\n");

	while (alpha <=maxQSize){
		/*------------------------------------------*/
		/* Measure time for UpDown simulation       */
		/*------------------------------------------*/
		
		printf("updown test alpha = %d, dist = %s\n", alpha, DistType);
		/*set up the distribution array with the correct distribution*/

		switch(k)
		{
		case 1: 
			SetTriangle(0,1,1, alpha , &myRand, myDist,maxSeconds);
			break;
		case 2:
			SetTriangle(0,1, 0, alpha, &myRand, myDist, maxSeconds);
			break;
		case 3:
			SetRectangle(alpha, &myRand, maxSeconds, myDist);
			break;
		case 4: 
			SetJitter(alpha, &myRand, maxSeconds, myDist);
			break;
		case 5: 
			SetCamel(alpha, 0.8, 0.2, &myRand, maxSeconds, myDist);
			break;
		case 6:
			SetCamel(alpha, 0.99, 0.01, &myRand, maxSeconds, myDist);
			break;
		
		}
	/*create n packets in the packet array*/
		createPackets(alpha, myDist);
	
		myCalendarQ = InitQueue();
		myCalendarQ->widthFactor = w;
			
		/*start = clock();*/
		QueryPerformanceCounter(&liStart);

		for (f = 0; f < alpha; f++){
			enqueue(myCalendarQ, packetList[f]);
		}
		for (i = 0; i < alpha; i++){
			packetList[i] = dequeue(myCalendarQ);
		}

		QueryPerformanceCounter(&liEnd);	
		
		
		/*finish = clock();*/
	
		duration = (double)((double)(liEnd.LowPart - liStart.LowPart) /((double) liPerfFreq.LowPart));
		durationPerQOperation  = duration / alpha;
	
		fprintf(fp, "%d     %1.9f \n",alpha, durationPerQOperation);
	
		alpha += delta;
	
		/*free the memory allocated to the calendar*/
		KillCalendar(myCalendarQ);
		fclose(fp);
		fopen("output.txt", "a");
	}
 }
 
 for (k = 1; k <= 6; k++){
	 alpha = delta;
	switch(k)

	{

	case 1:
		DistType = "Positive Triangular (0,1,1)";
		filename = "PosTriangle.txt";
		break;
	case 2:
		DistType = "Negative Triangular (0,1,0)";
		filename = "NegTriangle.txt";
		break;
	case 3:
		DistType = "Rectangular Distribution";
		filename = "Rectangle.txt";
		break;
	case 4:
		DistType = "Jitter Distribution";
		filename = "Jitter.txt";
		break;
	case 5:
		DistType = "Camel Distribution(0.8, 0.2)";
		filename = "Camel82.txt";
		break;
	case 6:
		DistType = "Camel Distribution(0.99, 0.01)";
		filename = "Camel9901.txt";
		break;
	}

	fprintf(fp, "#\n\n-------------------------------------------#\n");
	fprintf(fp, "\nDistribution: = %s \n", DistType);
	fprintf(fp, "#  Measure average time for the HOLD operation  #\n");
	fprintf(fp, "#-------------------------------------------#\n\n");
	fprintf(fp, "QSize     Average Hold	 Time To perform\n");
	fprintf(fp, "          time (seconds) %d Hold Operations\n", nHOLD);
	fprintf(fp, "______   _________________  ___________________\n");

	alpha =  delta;
	while (alpha <= maxQSize){
	
		/*----------------------------------------*/
		/*Measure average time for the HOLD operation */
		/*-----------------------------------------*/
		/*set up the distribution array with the correct distribution*/

	switch(k)
		{
		case 1: 
			SetTriangle(0,1,1, alpha + nHOLD , &myRand, myDist,maxSeconds);
			
			break;
		case 2:
			SetTriangle(0,1, 0, alpha + nHOLD, &myRand, myDist, maxSeconds);
			break;
		case 3:
			SetRectangle(alpha + nHOLD, &myRand, maxSeconds, myDist);
			break;
		case 4: 
			SetJitter(alpha + nHOLD, &myRand, maxSeconds, myDist);
			break;
		case 5: 
			SetCamel(alpha + nHOLD, 0.8, 0.2, &myRand, maxSeconds, myDist);
			break;
		case 6:
			SetCamel(alpha + nHOLD, 0.99, 0.01, &myRand, maxSeconds, myDist);
			break;
		
		}
	printf("Hold test alpha = %d, dist = %s\n", alpha, DistType);
		/*create n packets in the packet array*/

		createPackets(alpha, myDist);

	if ((alpha == 13000) && (k == 1)){
		fpDebug = fopen("debug.txt", "w");
		for (int x = 0; x < alpha; x++){
			fprintf(fpDebug, "%d   %lf\n", x, myDist->dataList[x]);
		}
		fclose(fpDebug);
	}
		/*initialize myCalendarQ*/
		myCalendarQ = InitQueue();
		myCalendarQ->widthFactor = w;
	
		for(i = 0; i < alpha; i++){
			enqueue(myCalendarQ, packetList[i]);
		} 

		/*print out distribution when Q size = 10000*/
		if (alpha == 10000){
			PrintEventDistribution(myCalendarQ, k, filename);}

		/*create nHolds Nodes*/
		next = alpha;
		/*start = clock();*/
		QueryPerformanceCounter(&liStart);
		for (j = 0; j < nHOLD; j++){
			myNode = dequeue(myCalendarQ);
			myNode->Prio = myDist->dataList[next + j];
			enqueue(myCalendarQ, myNode);	
		}
		QueryPerformanceCounter(&liEnd);
		/*	finish = clock();*/
		duration = (double)((double)(liEnd.LowPart - liStart.LowPart) /((double) liPerfFreq.LowPart));
		
		AverageHoldTime = (double)(duration/((double)nHOLD));

		fprintf(fp, "\n%d      %1.9f      %1.9f", alpha, AverageHoldTime, duration);
	
		fprintf(fp, "%d ", myCalendarQ->localinitCounter);
		alpha += delta;

		/*free all memory previously allocated to the calQ*/
		KillCalendar(myCalendarQ); 
		fclose(fp);
		fopen("output.txt", "a");
		}
	fprintf(fp, "\n\n");
	/*PrintEventDistribution(myCalendarQ);*/
 	fprintf(fp, "\n_____________________End of test %d ->%s______________\n\n", k, DistType);
	printf("\n end of test %d \n\n", k);
}
printf("end of program");
	

 return 0;
}
