/* queue.h*/
/* a dynamic calendar queue implementation*/
/* see Oh, SeungHyun and Ahn, JongSuk, 1999, "Dynamic Calendar Queue"*/

/*  Last Updated June 4/2001 */

#ifndef QUEUE_H
#define QUEUE_H

#include <stdlib.h>
#include <stdio.h>

#define TRUE 1
#define FALSE 0
#define QSPACE 40000

/*------------data structures--------------*/
/*Assumption: High priority values are numerically less than
  Low priority values*/

#define boolean int

typedef struct node node;
typedef struct Bucket Bucket;
typedef struct calendarQ calendarQ;


struct node{
  node* next;
  node* prev;
  double Prio;

};


 struct Bucket{
   double TopPrio; /*priority of element with highest priority in list */
  double BotPrio;
  node* head;
  node* tail;
  int nodeCount;
};


struct calendarQ{
   int qsize; /* # of packets in queue*/
   int nBuckets; /* number of buckets in queue */
   int top_threshold; /*max number of elements before resizing*/
  int bot_threshold; /*min number of elements before resizing*/
  double BucketWidth; /*length of a day in the calendar*/
   int LastBucket; /*bucket# in calendar from which last event was dequeued*/
  double BucketTop; /* Priority at top of LastBucket -the highest priority
		     that could go into that bucket (actually = to 1/2 width + top priority)*/
  double LastPrio; /* priority of last event dequeued*/
  
  Bucket* BcktContainer[QSPACE];      /*container for the bucket array (an array of pointers)
				-much larger than the array of buckets*/
  boolean resizeEnabled;      /*used to disable copying to a new calendar
				when temporarily dequeueing elements to 
				calculate a new bucket width */
  long firstsub;             /* beginning array index of BucketArray 
			       within the BcktContainer*/
	int SumEnqCnt;  /* the Sum of the number of visiting events for an enqueue operation*/
	int EnqSearch; /* number of enqueue operations after last resize*/
	int DeqSearch;
	int SumDeqCnt;

	long LargestBucket; /*the bucket with the largest event Q*/
	int resizeCount;
	float DCQThreshold;
	int DCQnumBuckets;
	int localinitCounter;
	int widthFactor;
	int newDCQWidthCounter;
};

/*------------Methods---------------------*/

calendarQ* InitQueue(void);
/*initializes an empty queue and returns a pointer to the new calendar queue structure*/

void LocalInit(calendarQ* myCalendarQ, int qbase, 
	       int nbuck, double bWidth, double startprio);
/*initializes a new calendar. Each new calendar generated by resize()
  is initialized by this procedure */
/* This initializes a bucket array within the array BcktContainer. Bucket
width is set equal to bwidth. BucketArray[0] is made equal to 
BcktContainer[qbase]; and the number of buckets is nbuck. Startprio
is the priority at which dequeueing begins. All external variables
except resizeenabled are initialized. */

void resize(calendarQ* Q,  int newBucketCount, int OptionSwitch, int eventSum);
/*this copies the queue onto a calendar with newBucketCount
buckets. The new bucket array is on the opposite end of the 
array BcktContainer[QSPACE] from the original */

Bucket* createBucket(double lBound, double uBound);

node* createNode(double priority);
/*returns a pointer to a node with the given priority*/

void enqueue(calendarQ* Q, node* entry);
/* adds one entry to the queue */

node* dequeue(calendarQ* Q);
/* removes the lowest priority node from the queue
  and returns a pointer to the node containing it.
also updates LastBucket and LastPrio just before 
returning a dequeued node */

double newDCQwidth(calendarQ* Q, int eventSum);

void DCQCheck_Distribution(calendarQ* Q);

double newwidth(calendarQ* Q);
/*this calculates the width to use for buckets in a new calendar. */


void printQ(calendarQ* Q);

/*prints out a current version of the Q*/


void fprintQ(calendarQ* Q, FILE *fp );
/*prints out a current version of the Q to a file*/



#endif
